import os
import sys
import json
import requests
from urllib.parse import urlparse

from metaflow import current
from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow.plugins.parallel_decorator import ParallelDecorator
from metaflow.metadata_provider.util import sync_local_metadata_to_datastore
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.sidecar import Sidecar
from metaflow.plugins.timeout_decorator import get_run_time_limit_for_task
from metaflow.metadata_provider import MetaDatum
from metaflow.metaflow_config_funcs import init_config
from .constants import SUPPORTABLE_GPU_TYPES, DEFAULT_GPU_TYPE
from .exceptions import (
    RequestedGPUTypeUnavailableException,
    UnsupportedNvcfConfigurationException,
    UnsupportedNvcfDatastoreException,
    NvcfTimeoutTooShortException,
    NvcfQueueTimeoutTooShortException,
)

from metaflow.metaflow_config import SERVICE_URL


class NvcfDecorator(StepDecorator):

    """
    Specifies that this step should execute on DGX cloud.

    Parameters
    ----------
    gpu : int
        Number of GPUs to use.
    gpu_type : str
        Type of Nvidia GPU to use.
    queue_timeout : int
        Time to keep the job in NVCF's queue.
    """

    name = "nvidia"
    defaults = {
        "gpu": 1,
        "gpu_type": None,
        "queue_timeout": 5 * 24 * 3600,  # Default 5 days in seconds
    }

    package_url = None
    package_sha = None

    # Refer https://github.com/Netflix/metaflow/blob/master/docs/lifecycle.png
    # to understand where these functions are invoked in the lifecycle of a
    # Metaflow flow.
    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # Executing NVCF functions requires a non-local datastore.
        if flow_datastore.TYPE not in ("s3", "azure", "gs"):
            raise UnsupportedNvcfDatastoreException(flow_datastore.TYPE)

        # Set internal state.
        self.logger = logger
        self.environment = environment
        self.step = step
        self.flow_datastore = flow_datastore

        if any([deco.name == "kubernetes" for deco in decos]):
            raise MetaflowException(
                "Step *{step}* is marked for execution both on Kubernetes and "
                "Nvidia. Please use one or the other.".format(step=step)
            )
        if any([isinstance(deco, ParallelDecorator) for deco in decos]):
            raise MetaflowException(
                "Step *{step}* contains a @parallel decorator "
                "with the @nvidia decorator. @parallel decorators are not currently supported with @nvidia.".format(
                    step=step
                )
            )

        # Set run time limit for the NVCF function.
        self.run_time_limit = get_run_time_limit_for_task(decos)
        if self.run_time_limit < 60:
            raise NvcfTimeoutTooShortException(step)

        conf = init_config()
        if "OBP_AUTH_SERVER" in conf:
            auth_host = conf["OBP_AUTH_SERVER"]
        else:
            auth_host = "auth." + urlparse(SERVICE_URL).hostname.split(".", 1)[1]

        # NOTE: reusing the same auth_host as the one used in NimMetadata,
        # however, user should not need to use nim container to use @nvidia.
        # May want to refactor this to a common endpoint.
        nim_info_url = "https://" + auth_host + "/generate/nim"

        if "METAFLOW_SERVICE_AUTH_KEY" in conf:
            headers = {"x-api-key": conf["METAFLOW_SERVICE_AUTH_KEY"]}
            res = requests.get(nim_info_url, headers=headers)
        else:
            headers = json.loads(os.environ.get("METAFLOW_SERVICE_HEADERS"))
            res = requests.get(nim_info_url, headers=headers)

        res.raise_for_status()
        self.attributes["ngc_api_key"] = res.json()["nvcf"]["api_key"]

        available_functions_info = res.json()["nvcf"]["functions"]
        requested_gpu_type = self.attributes["gpu_type"]
        n_gpu = self.attributes["gpu"]

        if requested_gpu_type is None:
            requested_gpu_type = DEFAULT_GPU_TYPE
        if requested_gpu_type not in SUPPORTABLE_GPU_TYPES:
            raise RequestedGPUTypeUnavailableException(requested_gpu_type)

        desired_configuration = (n_gpu, requested_gpu_type)
        available_configurations = {}
        for f in available_functions_info:
            if f["model_key"] == "metaflow_task_executor":
                available_configurations[(f["gpu"], f["gpu_type"])] = f["id"]

        if desired_configuration not in available_configurations:
            raise UnsupportedNvcfConfigurationException(
                n_gpu, requested_gpu_type, available_configurations, step
            )
        self.attributes["function_id"] = available_configurations[desired_configuration]

        queue_timeout = self.attributes["queue_timeout"]
        if not isinstance(queue_timeout, int) or queue_timeout < 60:
            raise NvcfQueueTimeoutTooShortException(step)

    def runtime_init(self, flow, graph, package, run_id):
        # Set some more internal state.
        self.flow = flow
        self.graph = graph
        self.package = package
        self.run_id = run_id

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        if not is_cloned:
            self._save_package_once(self.flow_datastore, self.package)

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        if retry_count <= max_user_code_retries:
            # after all attempts to run the user code have failed, we don't need
            # to execute on NVCF anymore. We can execute possible fallback
            # code locally.
            cli_args.commands = ["nvidia", "step"]
            cli_args.command_args.append(self.package_sha)
            cli_args.command_args.append(self.package_url)
            cli_options = {
                "function_id": self.attributes["function_id"],
                "ngc_api_key": self.attributes["ngc_api_key"],
                "queue_timeout": self.attributes["queue_timeout"],
            }
            cli_args.command_options.update(cli_options)
            cli_args.entrypoint[0] = sys.executable

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_retries,
        ubf_context,
        inputs,
    ):
        self.metadata = metadata
        self.task_datastore = task_datastore

        # task_pre_step may run locally if fallback is activated for @catch
        # decorator.

        if "NVCF_CONTEXT" in os.environ:
            meta = {}

            meta["nvcf-function-id"] = os.environ.get("NVCF_FUNCTION_ID")
            meta["nvcf-function-version-id"] = os.environ.get(
                "NVCF_FUNCTION_VERSION_ID"
            )
            meta["nvcf-region"] = os.environ.get("NVCF_REGION")
            meta["nvcf-ncaid"] = os.environ.get("NVCF_NCAID")
            meta["nvcf-sub"] = os.environ.get("NVCF_SUB")
            meta["nvcf-instancetype"] = os.environ.get("NVCF_INSTANCETYPE")
            meta["nvcf-reqid"] = os.environ.get("NVCF_REQID")
            meta["nvcf-env"] = os.environ.get("NVCF_ENV")
            meta["nvcf-backend"] = os.environ.get("NVCF_BACKEND")
            meta["nvcf-function-name"] = os.environ.get("NVCF_FUNCTION_NAME")
            meta["nvcf-nspectid"] = os.environ.get("NVCF_NSPECTID")

            entries = [
                MetaDatum(field=k, value=v, type=k, tags=[])
                for k, v in meta.items()
                if v is not None
            ]
            # Register book-keeping metadata for debugging.
            metadata.register_metadata(run_id, step_name, task_id, entries)

            self._save_logs_sidecar = Sidecar("save_logs_periodically")
            self._save_logs_sidecar.start()

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_retries
    ):
        # task_finished may run locally if fallback is activated for @catch
        # decorator.
        if "NVCF_CONTEXT" in os.environ:
            # If `local` metadata is configured, we would need to copy task
            # execution metadata from the NVCF container to user's
            # local file system after the user code has finished execution.
            # This happens via datastore as a communication bridge.
            if hasattr(self, "metadata") and self.metadata.TYPE == "local":
                sync_local_metadata_to_datastore(
                    DATASTORE_LOCAL_DIR, self.task_datastore
                )

        try:
            self._save_logs_sidecar.terminate()
        except:
            # Best effort kill
            pass

    @classmethod
    def _save_package_once(cls, flow_datastore, package):
        if cls.package_url is None:
            cls.package_url, cls.package_sha = flow_datastore.save_data(
                [package.blob], len_hint=1
            )[0]



import json
import os
import time
import threading
from urllib.request import HTTPError, Request, URLError, urlopen
from functools import wraps

from metaflow import util
from metaflow.mflog import (
    BASH_SAVE_LOGS,
    bash_capture_logs,
    export_mflog_env_vars,
    tail_logs,
    get_log_tailer,
)
from .exceptions import NvcfJobFailedException, NvcfPollingConnectionError

# Redirect structured logs to $PWD/.logs/
LOGS_DIR = "$PWD/.logs"
STDOUT_FILE = "mflog_stdout"
STDERR_FILE = "mflog_stderr"
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)


def retry_on_status(status_codes=[500], max_retries=3, delay=1):
    def decorator(func):
        @wraps(func)
        def wrapper(instance, *args, **kwargs):
            retries = 0

            # Determine retry limit upfront
            use_queue_timeout = 504 in status_codes
            if use_queue_timeout:
                poll_seconds = int(instance._poll_seconds)
                retry_limit = (
                    instance._queue_timeout + (poll_seconds - 1)
                ) // poll_seconds
                remainder = instance._queue_timeout % poll_seconds
                last_timeout = remainder if remainder != 0 else poll_seconds
            else:
                retry_limit = max_retries

            while retries < retry_limit:
                try:
                    return func(instance, *args, **kwargs)
                except HTTPError as e:
                    if e.code not in status_codes or retries >= retry_limit:
                        instance._status = JobStatus.FAILED
                        if e.code == 504 and retries >= retry_limit:
                            raise NvcfPollingConnectionError(
                                "Request timed out after all retries"
                            )
                        raise

                    if e.code == 504 and retries == retry_limit - 1:
                        instance._poll_seconds = str(last_timeout)

                    print(
                        f"[@nvidia] {'Queue timeout' if e.code == 504 else f'Received {e.code}'}, "
                        f"retrying ({retries + 1}/{retry_limit})... with poll seconds as {instance._poll_seconds}"
                    )

                    if e.code != 504:
                        time.sleep(delay)

                    retries += 1
                except URLError as e:
                    instance._status = JobStatus.FAILED
                    raise
            # final attempt
            return func(instance, *args, **kwargs)

        return wrapper

    return decorator


class Nvcf(object):
    def __init__(
        self, metadata, datastore, environment, function_id, ngc_api_key, queue_timeout
    ):
        self.metadata = metadata
        self.datastore = datastore
        self.environment = environment
        self._function_id = function_id
        self._ngc_api_key = ngc_api_key
        self._queue_timeout = queue_timeout

    def launch_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        env={},
    ):
        mflog_expr = export_mflog_env_vars(
            datastore_type=code_package_ds,
            stdout_path=STDOUT_PATH,
            stderr_path=STDERR_PATH,
            **task_spec,
        )
        init_cmds = self.environment.get_package_commands(
            code_package_url, code_package_ds
        )
        init_expr = " && ".join(init_cmds)
        heartbeat_expr = f'python -m metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store "$MAIN_PID" {code_package_ds} nvcf_heartbeats & HEARTBEAT_PID=$!;'
        step_expr = bash_capture_logs(
            " && ".join(
                self.environment.bootstrap_commands(step_name, code_package_ds)
                + [step_cli + " & MAIN_PID=$!; " + heartbeat_expr + " wait $MAIN_PID"]
            )
        )

        # construct an entry point that
        # 1) initializes the mflog environment (mflog_expr)
        # 2) bootstraps a metaflow environment (init_expr)
        # 3) executes a task (step_expr)

        cmd_str = "mkdir -p %s && %s && %s && %s; " % (
            LOGS_DIR,
            mflog_expr,
            init_expr,
            step_expr,
        )
        # after the task has finished, we save its exit code (fail/success)
        # and persist the final logs. The whole entrypoint should exit
        # with the exit code (c) of the task.
        #
        # Note that if step_expr OOMs, this tail expression is never executed.
        # We lose the last logs in this scenario.
        cmd_str += (
            "c=$?; kill $HEARTBEAT_PID; wait $HEARTBEAT_PID; %s; exit $c"
            % BASH_SAVE_LOGS
        )
        cmd_str = (
            '${METAFLOW_INIT_SCRIPT:+eval \\"${METAFLOW_INIT_SCRIPT}\\"} && %s'
            % cmd_str
        )
        self.job = Job(
            'bash -c "%s"' % cmd_str,
            env,
            task_spec,
            self.datastore._storage_impl,
            self._function_id,
            self._ngc_api_key,
            self._queue_timeout,
        )
        self.job.submit()

    def wait(self, stdout_location, stderr_location, echo=None):
        def wait_for_launch(job):
            status = job.status
            echo(
                "Task status: %s..." % status,
                "stderr",
                _id=job.id,
            )

        prefix = b"[%s] " % util.to_bytes(self.job.id)
        stdout_tail = get_log_tailer(stdout_location, self.datastore.TYPE)
        stderr_tail = get_log_tailer(stderr_location, self.datastore.TYPE)

        # 1) Loop until the job has started
        wait_for_launch(self.job)

        # 2) Tail logs until the job has finished
        tail_logs(
            prefix=prefix,
            stdout_tail=stdout_tail,
            stderr_tail=stderr_tail,
            echo=echo,
            has_log_updates=lambda: self.job.is_running,
        )

        echo(
            "Task finished with exit code %s." % self.job.result.get("exit_code"),
            "stderr",
            _id=self.job.id,
        )
        if self.job.has_failed:
            raise NvcfJobFailedException(
                "This could be a transient error. Use @retry to retry."
            )


class JobStatus(object):
    CREATED = "CREATED"  # Job object created but not submitted
    SUBMITTED = "SUBMITTED"  # Job submitted to NVCF
    POLLED = "POLLED"  # Job has been successfully polled at least once
    SUCCESSFUL = "SUCCESSFUL"  # Job completed successfully
    FAILED = "FAILED"  # Job failed
    DISAPPEARED = "DISAPPEARED"  # Job disappeared from NVCF but was previously polled (likely successful)


nvcf_url = "https://api.nvcf.nvidia.com"
submit_endpoint = f"{nvcf_url}/v2/nvcf/pexec/functions"
result_endpoint = f"{nvcf_url}/v2/nvcf/pexec/status"


class Job(object):
    def __init__(
        self, command, env, task_spec, backend, function_id, ngc_api_key, queue_timeout
    ):
        self._payload = {
            "command": command,
            "env": {k: v for k, v in env.items() if v is not None},
        }
        self._result = {}
        self._function_id = function_id
        self._ngc_api_key = ngc_api_key
        self._queue_timeout = queue_timeout
        self._poll_seconds = "3600"

        # Initialize status and tracking variables
        self._status = JobStatus.CREATED
        self._last_poll_time = time.time()
        self._force_poll_interval = 30

        flow_name = task_spec.get("flow_name")
        run_id = task_spec.get("run_id")
        step_name = task_spec.get("step_name")
        task_id = task_spec.get("task_id")
        retry_count = task_spec.get("retry_count")

        heartbeat_prefix = "/".join(
            (flow_name, str(run_id), step_name, str(task_id), str(retry_count))
        )

        ## import is done here to avoid the following warning:
        # RuntimeWarning: 'metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store' found in sys.modules
        # after import of package 'metaflow_extensions.outerbounds.plugins.nvcf', but prior to execution of
        # 'metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store'; this may result in unpredictable behaviour
        from metaflow_extensions.outerbounds.plugins.nvcf.heartbeat_store import (
            HeartbeatStore,
        )

        store = HeartbeatStore(
            main_pid=None,
            storage_backend=backend,
        )

        self.heartbeat_thread = threading.Thread(
            target=store.emit_heartbeat,
            args=(
                heartbeat_prefix,
                "nvcf_heartbeats",
            ),
            daemon=True,
        )
        self.heartbeat_thread.start()

    @retry_on_status(status_codes=[504])
    def submit(self):
        try:
            headers = {
                "Authorization": f"Bearer {self._ngc_api_key}",
                "Content-Type": "application/json",
                "nvcf-feature-enable-gateway-timeout": "true",
                "NVCF-POLL-SECONDS": self._poll_seconds,
            }
            request_data = json.dumps(self._payload).encode()
            request = Request(
                f"{submit_endpoint}/{self._function_id}",
                data=request_data,
                headers=headers,
            )
            response = urlopen(request)
            self._invocation_id = response.headers.get("NVCF-REQID")
            if response.getcode() == 200:
                data = json.loads(response.read())
                if data.get("exit_code") == 0:
                    self._status = JobStatus.SUCCESSFUL
                else:
                    self._status = JobStatus.FAILED
                self._result = data
            elif response.getcode() == 202:
                self._status = JobStatus.SUBMITTED
            else:
                self._status = JobStatus.FAILED
        except URLError:
            self._status = JobStatus.FAILED
            raise

    @property
    def status(self):
        terminal_states = [
            JobStatus.SUCCESSFUL,
            JobStatus.FAILED,
            JobStatus.DISAPPEARED,
        ]

        # If status is already terminal, don't poll again
        if self._status in terminal_states:
            return self._status

        current_time = time.time()
        if (
            current_time - self._last_poll_time > self._force_poll_interval
            or self._status not in terminal_states
        ):
            try:
                self._poll()
                self._last_poll_time = current_time

                # Update job status to POLLED if this is our first successful poll
                if self._status == JobStatus.SUBMITTED:
                    self._status = JobStatus.POLLED

                if self._status == JobStatus.SUCCESSFUL:
                    return self._status

            except HTTPError as e:
                if e.code == 404:
                    # 404 interpretation depends on job lifecycle
                    if self._status in [JobStatus.POLLED, JobStatus.SUBMITTED]:
                        # We've submitted or successfully polled this job before,
                        # so a 404 likely means it completed and was removed
                        self._status = JobStatus.DISAPPEARED
                        self._result = {"exit_code": 0}
                    else:
                        # Job was never successfully tracked
                        print(
                            f"[@nvidia] 404 received for job that was never successfully tracked - treating as failure"
                        )
                        self._status = JobStatus.FAILED
                        raise NvcfPollingConnectionError(e)
                else:
                    self._status = JobStatus.FAILED
                    raise NvcfPollingConnectionError(e)
            except URLError as e:
                self._status = JobStatus.FAILED
                raise NvcfPollingConnectionError(e)
        return self._status

    @property
    def id(self):
        return self._invocation_id

    @property
    def is_running(self):
        # Job is running if it's in SUBMITTED or POLLED state
        return self.status in [JobStatus.SUBMITTED, JobStatus.POLLED]

    @property
    def has_failed(self):
        return self.status == JobStatus.FAILED

    @property
    def result(self):
        return self._result

    @retry_on_status(status_codes=[500], max_retries=3, delay=5)
    @retry_on_status(status_codes=[504])
    def _poll(self):
        try:
            headers = {
                "Authorization": f"Bearer {self._ngc_api_key}",
                "Content-Type": "application/json",
                "nvcf-feature-enable-gateway-timeout": "true",
                "NVCF-POLL-SECONDS": self._poll_seconds,
            }
            request = Request(
                f"{result_endpoint}/{self._invocation_id}", headers=headers
            )
            response = urlopen(request)
            body = response.read()
            print(f"[@nvidia] polling response: {body}")
            if response.getcode() == 200:
                data = json.loads(body)
                # TODO: Propagate the internal error forward
                if data.get("exit_code") == 0:
                    self._status = JobStatus.SUCCESSFUL
                else:
                    self._status = JobStatus.FAILED
                self._result = data
            elif response.getcode() != 202:
                print(
                    f"[@nvidia] Unexpected response code: {response.getcode()}. Please notify an Outerbounds support engineer if this error persists."
                )
                self._status = JobStatus.FAILED
        except URLError:
            self._status = JobStatus.FAILED
            raise



from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow import current
from .app_utils import start_app
from .supervisord_utils import SupervisorClient, SupervisorClientException
import os
import random
import string
import tempfile
import sys
from .consts import (
    BASE_DIR_FOR_APP_ASSETS,
    DEFAULT_WAIT_TIME_SECONDS_FOR_PROCESS_TO_START,
)

"""
There are 3 variants of starting apps that we support through this function. Which variant is applied is
a result of the user setting (or not setting) self.entrypoint and self.app_dir.

The chosen variant determines whether or not the user will have an easy (auto-magical) way to write their
Metaflow artifacts the the directory where their app will be run from. This will simplify managing models
from inside the code.

Case 1:
    Desired Behavior:
        The user doesn't care about auto-magical artifact management, they just have a server.py somewhere and
        they want to run it.
    How:
        The user sets self.entrypoint to the name of their file (optionally with any args).
        The value of self.app_dir is irrelevant.
        Example: self.entrypoint = "/home/ob-workspace/my_random_directory/my_subfolder/server.py --my_arg 764"
    Expected Behavior:
        The users app will be started using a conda environment built by Metaflow, that's it.

Case 2:
    Desired Behavior:
        The user has defined a clean package/module structure with a __main__.py in a folder somewhere. They would like
        to run this module as an app. They would like to access models in the same top level directory as their app.
    How:
        The user sets self.entrypoint to None.
        The user sets self.app_dir to the top level directory of their app.
        Example: self.app_dir = "/home/ob-workspace/my_random_directory/my_subfolder"
        (my_subfolder HAS to contain a __main__.py file.)
    Expected Behavior:
        The users app will be started using a conda environment built by Metaflow. The user gets access to self.deploy_dir where they
        write their artifacts. After the user writes their artifacts to self.deploy_dir, the artifacts and the app is copied over
        to an internal directory where we will deploy the app from.
        The internal directory is: /home/ob-workspace/.appdaemon/apps/<app_port>.
        <app_port> will contain user's __main__.py and other files (copied recursively), as well as their artifacts that they wrote to
        self.deploy_dir. self.deploy_dir/my_model becomes <app_port>/my_model. my_subfolder/__main__.py becomes <app_port>/__main__.py.

Case 3:
    Desired Behavior:
        The user follows the Outerbounds convention, which is the same as Case 2, except the app is actually in the same top level folder
        as the Deployer flow. They would like to access models as usual.
    How:
        The user sets self.entrypoint to None (or doesn't set it at all).
        The user sets self.app_dir to None (or doesn't set it at all).
    Expected Behavior:
        The users app will be started using a conda environment built by Metaflow. Everything else is exactly the same as Case 2.
"""


class WorkstationAppDeployDecorator(StepDecorator):
    """
    Specifies that this step is used to deploy an instance of the app.
    Requires that self.app_name, self.app_port, self.entrypoint and self.deployDir is set.

    Parameters
    ----------
    app_port : int
        Number of GPUs to use.
    app_name : str
        Name of the app to deploy.
    """

    name = "app_deploy"
    defaults = {"app_port": 8080, "app_name": "app"}

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        if any([deco.name == "kubernetes" for deco in decos]):
            raise MetaflowException(
                "@app_deploy decorator is only supported locally and does not work with remote execution environments like @kubernetes, @nvidia."
            )

        # We always need to have some environment defined through the flow to deploy and app.
        # Which means either step decorators like @pypi / @conda must be defined.
        # or flow level decorators like @conda_base / @pypi_base.
        if not any([deco.name == "pypi" or deco.name == "conda" for deco in decos]):
            flow_decorators = flow._flow_decorators.keys()
            if (
                "conda_base" not in flow_decorators
                and "pypi_base" not in flow_decorators
            ):
                raise MetaflowException(
                    "@app_deploy requires either step decorators like @pypi / @conda or flow level decorators like @conda_base / @pypi_base to be defined."
                )

        app_port = self.attributes["app_port"]
        app_name = self.attributes["app_name"]

        # Currently this decorator is expected to only execute on workstation.
        if app_port is None or app_port < 6000 or app_port > 6002:
            raise MetaflowException(
                "AppDeployDecorator requires app_port to be between 6000 and 6002."
            )

        if app_name is None:
            raise MetaflowException("AppDeployDecorator requires app_name to be set.")

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        """
        Runs before the step decorated with @app_deploy has started.
        We create a directory where the user can write their artifacts and expose it via self.deploy_dir.
        """
        os.makedirs(BASE_DIR_FOR_APP_ASSETS, exist_ok=True)
        # First we want to create a directory where the user's app directory and artifacts can be stored.
        with tempfile.TemporaryDirectory(
            prefix=BASE_DIR_FOR_APP_ASSETS, delete=False
        ) as temp_dir:
            launch_temp_dir = temp_dir

        # Expose this to the user, so that they can use it write their artifacts.
        setattr(flow, "deploy_dir", launch_temp_dir)

        # Make sure to record deploy_dir so that the user cannot accidentally override it.
        self._deploy_dir = launch_temp_dir

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        """
        Runs after the step decorated with @app_deploy has finished.
        Based on the cases above, things that we care about are self.entrypoint and self.app_dir.
        """

        deploy_dir = self._deploy_dir

        # By default we assume that the user has a __main__.py file in their app directory.
        # They can always override this behavior.
        user_provided_entrypoint = getattr(flow, "entrypoint", None)

        if user_provided_entrypoint is not None and not isinstance(
            user_provided_entrypoint, str
        ):
            raise MetaflowException(
                f"@app_deploy requires entrypoint to be set to a string. The current value of entrypoint {user_provided_entrypoint} is not valid."
            )

        flow_directory = os.path.dirname(os.path.abspath(sys.argv[0]))

        app_location = getattr(
            flow, "app_dir", os.path.join(flow_directory, self.attributes["app_name"])
        )

        if user_provided_entrypoint is None and not os.path.exists(app_location):
            raise MetaflowException(f"App directory {app_location} does not exist.")

        wait_time_for_app_start = getattr(
            flow,
            "wait_time_for_app_start",
            DEFAULT_WAIT_TIME_SECONDS_FOR_PROCESS_TO_START,
        )

        try:
            supervisor_client = SupervisorClient(
                wait_time_seconds_for_app_start=wait_time_for_app_start
            )

            # First, let's deploy the app.
            start_app(
                port=self.attributes["app_port"], name=self.attributes["app_name"]
            )

            # Now, let's add the app to supervisor.
            supervisor_client.start_process_with_supervisord(
                self.attributes["app_name"],
                self.attributes["app_port"],
                user_provided_entrypoint,
                deploy_dir,
                app_location,
            )
        except SupervisorClientException as e:
            raise MetaflowException(str(e))
        except Exception as e:
            raise MetaflowException(
                f"Failed to start {self.attributes['app_name']}! Cause: {str(e)}"
            ) from e



import os, sys, time, json, random, requests, sqlite3
from urllib.parse import urlparse
from metaflow.metaflow_config import SERVICE_URL
from metaflow.metaflow_config_funcs import init_config
from .utilities import get_storage_path
from ..nvcf.nvcf import retry_on_status


NVCF_URL = "https://api.nvcf.nvidia.com"
NVCF_SUBMIT_ENDPOINT = f"{NVCF_URL}/v2/nvcf/pexec/functions"
NVCF_RESULT_ENDPOINT = f"{NVCF_URL}/v2/nvcf/pexec/status"
NVCF_POLL_INTERVAL_SECONDS = 1
COMMON_HEADERS = {
    "accept": "application/json",
    "Content-Type": "application/json",
    "nvcf-feature-enable-gateway-timeout": "true",
}


class NimMetadata(object):
    def __init__(self):
        self._nvcf_chat_completion_models = []
        self._coreweave_chat_completion_models = []

        conf = init_config()

        if "OBP_AUTH_SERVER" in conf:
            auth_host = conf["OBP_AUTH_SERVER"]
        else:
            auth_host = "auth." + urlparse(SERVICE_URL).hostname.split(".", 1)[1]

        nim_info_url = "https://" + auth_host + "/generate/nim"

        if "METAFLOW_SERVICE_AUTH_KEY" in conf:
            headers = {"x-api-key": conf["METAFLOW_SERVICE_AUTH_KEY"]}
            res = requests.get(nim_info_url, headers=headers)
        else:
            headers = json.loads(os.environ.get("METAFLOW_SERVICE_HEADERS"))
            res = requests.get(nim_info_url, headers=headers)

        res.raise_for_status()
        self._ngc_api_key = res.json()["nvcf"]["api_key"]

        for model in res.json()["nvcf"]["functions"]:
            self._nvcf_chat_completion_models.append(
                {
                    "name": model["model_key"],
                    "function-id": model["id"],
                    "version-id": model["version"],
                }
            )
        for model in res.json()["coreweave"]["containers"]:
            self._coreweave_chat_completion_models.append(
                {"name": model["nim_name"], "ip-address": model["ip_addr"]}
            )

    def get_nvcf_chat_completion_models(self):
        return self._nvcf_chat_completion_models

    def get_headers_for_nvcf_request(self):
        return {**COMMON_HEADERS, "Authorization": f"Bearer {self._ngc_api_key}"}


class NimManager(object):
    def __init__(self, models, backend, flow, step_name, monitor, queue_timeout):

        nim_metadata = NimMetadata()
        if backend == "managed":
            nvcf_models = [
                m["name"] for m in nim_metadata.get_nvcf_chat_completion_models()
            ]

            self.models = {}
            for m in models:
                if m in nvcf_models:
                    self.models[m] = NimChatCompletion(
                        model=m,
                        provider="NVCF",
                        nim_metadata=nim_metadata,
                        monitor=monitor,
                        queue_timeout=queue_timeout,
                    )
                else:
                    raise ValueError(
                        f"Model {m} not supported by the Outerbounds @nim offering."
                        f"\nYou can choose from these options: {nvcf_models}\n\n"
                        "Reach out to Outerbounds if there are other models you'd like supported."
                    )
        else:
            raise ValueError(
                f"Backend {backend} not supported by the Outerbounds @nim offering. Please reach out to Outerbounds."
            )


class JobStatus(object):
    SUBMITTED = "SUBMITTED"
    RUNNING = "RUNNING"
    SUCCESSFUL = "SUCCESSFUL"
    FAILED = "FAILED"


class NimChatCompletion(object):
    def __init__(
        self,
        model="meta/llama3-8b-instruct",
        provider="NVCF",
        nim_metadata=None,
        monitor=False,
        queue_timeout=None,
        **kwargs,
    ):
        if nim_metadata is None:
            raise ValueError(
                "NimMetadata object is required to initialize NimChatCompletion object."
            )

        self._nim_metadata = nim_metadata
        self.compute_provider = provider
        self.invocations = []
        self.max_request_retries = int(
            os.environ.get("METAFLOW_EXT_HTTP_MAX_RETRIES", "10")
        )
        self.monitor = monitor

        if self.compute_provider == "NVCF":
            nvcf_model_names = [
                m["name"] for m in self._nim_metadata.get_nvcf_chat_completion_models()
            ]
            self.model = model
            self.function_id = self._nim_metadata.get_nvcf_chat_completion_models()[
                nvcf_model_names.index(model)
            ]["function-id"]
            self.version_id = self._nim_metadata.get_nvcf_chat_completion_models()[
                nvcf_model_names.index(model)
            ]["version-id"]
        else:
            raise ValueError(
                f"Backend compute provider {self.compute_provider} not yet supported for @nim."
            )

        # to know whether to set file_name
        self.first_request = True

        # TODO (Eddie) - this may make more sense in a base class.
        # @nim arch needs redesign if customers start using it in more creative ways.
        self._poll_seconds = "3600"
        self._queue_timeout = queue_timeout
        self._status = None
        self._result = {}

    @property
    def status(self):
        return self._status

    @property
    def has_failed(self):
        return self._status == JobStatus.FAILED

    @property
    def is_running(self):
        return self._status == JobStatus.SUBMITTED

    @property
    def result(self):
        return self._result

    def _log_stats(self, response, e2e_time):
        stats = {}
        if response.status_code == 200:
            stats["success"] = 1
            stats["error"] = 0
        else:
            stats["success"] = 0
            stats["error"] = 1
        stats["status_code"] = response.status_code
        try:
            stats["prompt_tokens"] = response.json()["usage"]["prompt_tokens"]
        except KeyError:
            stats["prompt_tokens"] = None
        try:
            stats["completion_tokens"] = response.json()["usage"]["completion_tokens"]
        except KeyError:
            stats["completion_tokens"] = None
        stats["e2e_time"] = e2e_time
        stats["provider"] = self.compute_provider
        stats["model"] = self.model

        conn = sqlite3.connect(self.file_name)
        cursor = conn.cursor()
        try:
            cursor.execute(
                """
                INSERT INTO metrics (error, success, status_code, prompt_tokens, completion_tokens, e2e_time, model)
                VALUES (?, ?, ?, ?, ?, ?, ?)
            """,
                (
                    stats["error"],
                    stats["success"],
                    stats["status_code"],
                    stats["prompt_tokens"],
                    stats["completion_tokens"],
                    stats["e2e_time"],
                    stats["model"],
                ),
            )
            conn.commit()
        finally:
            conn.close()

    @retry_on_status(status_codes=[500], max_retries=3, delay=5)
    @retry_on_status(status_codes=[504])
    def __call__(self, **kwargs):

        if self.first_request:
            # Put here to guarantee self.file_name is set after task_id exists.
            from metaflow import current

            self.file_name = get_storage_path(current.task_id)

        request_data = {"model": self.model, **kwargs}
        request_url = f"{NVCF_SUBMIT_ENDPOINT}/{self.function_id}"
        retry_delay = 1
        attempts = 0
        t0 = time.time()
        while attempts < self.max_request_retries:
            try:
                attempts += 1
                response = requests.post(
                    request_url,
                    headers=self._nim_metadata.get_headers_for_nvcf_request(),
                    json=request_data,
                )
                if response.status_code == 202:
                    invocation_id = response.headers.get("NVCF-REQID")
                    self.invocations.append(invocation_id)
                    self._status = JobStatus.SUBMITTED
                elif response.status_code == 200:
                    tf = time.time()
                    if self.monitor:
                        self._log_stats(response, tf - t0)
                    self._status = JobStatus.SUCCESSFUL
                    self._result = response.json()
                    return self._result
                elif response.status_code == 400:
                    self._status = JobStatus.FAILED
                    msg = (
                        "[@nim ERROR] The OpenAI-compatible returned a 400 status code. "
                        + "Known causes include improper requests or prompts with too many tokens for the selected model. "
                        + "Please contact Outerbounds if you need assistance resolving the issue."
                    )
                    print(msg, file=sys.stderr)
                    self._result = {"ERROR": msg}
                    return self._result
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ) as e:
                # ConnectionErrors are generally temporary errors like DNS resolution failures,
                # timeouts etc.
                print(
                    "received error of type {}. Retrying...".format(type(e)),
                    e,
                    file=sys.stderr,
                )
                time.sleep(retry_delay)
                retry_delay *= 2  # Double the delay for the next attempt
                retry_delay += random.uniform(0, 1)  # Add jitter
                retry_delay = min(retry_delay, 10)

        def _poll():
            poll_request_url = f"{NVCF_RESULT_ENDPOINT}/{invocation_id}"
            attempts = 0
            retry_delay = 1
            while attempts < self.max_request_retries:
                try:
                    attempts += 1
                    poll_response = requests.get(
                        poll_request_url,
                        headers=self._nim_metadata.get_headers_for_nvcf_request(),
                    )
                    if poll_response.status_code == 200:
                        tf = time.time()
                        self._log_stats(response, tf - t0)
                        self._status = JobStatus.SUCCESSFUL
                        self._result = poll_response.json()
                        return self._result
                    elif poll_response.status_code == 202:
                        self._status = JobStatus.SUBMITTED
                        return 202
                    elif poll_response.status_code == 400:
                        self._status = JobStatus.FAILED
                        msg = (
                            "[@nim ERROR] The OpenAI-compatible API returned a 400 status code. "
                            + "Known causes include improper requests or prompts with too many tokens for the selected model. "
                            + "Please contact Outerbounds if you need assistance resolving the issue."
                        )
                        print(msg, file=sys.stderr)
                        self._result = {"@nim ERROR": msg}
                        return self._result
                except (
                    requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                ) as e:
                    print(
                        "received error of type {}. Retrying...".format(type(e)),
                        e,
                        file=sys.stderr,
                    )
                    time.sleep(retry_delay)
                    retry_delay *= 2  # Double the delay for the next attempt
                    retry_delay += random.uniform(0, 1)  # Add jitter
                    retry_delay = min(retry_delay, 10)

        while True:
            data = _poll()
            if data and data != 202:
                return data
            time.sleep(NVCF_POLL_INTERVAL_SECONDS)



import sqlite3
from metaflow.cards import Markdown, Table
from metaflow.metaflow_current import current

from .utilities import get_storage_path
from ..card_utilities.async_cards import CardRefresher
from ..card_utilities.extra_components import BarPlot, ViolinPlot


class NimMetricsRefresher(CardRefresher):
    CARD_ID = "nim_metrics"

    def __init__(self) -> None:
        self._metrics_charts = {}
        self._last_updated_on = None
        self._already_rendered = False
        self._file_name = get_storage_path(current.task_id)

    def sqlite_fetch_func(self, conn):
        cursor = conn.cursor()
        try:
            conn = sqlite3.connect(self._file_name)
            cursor = conn.cursor()
            cursor.execute(
                "SELECT error, success, status_code, prompt_tokens, completion_tokens, e2e_time, model FROM metrics"
            )
            rows = cursor.fetchall()
            data = {
                "error": 0,
                "success": 0,
                "status_code": [],
                "prompt_tokens": [],
                "completion_tokens": [],
                "e2e_time": [],
                "model": [],
            }
            for row in rows:
                data["error"] += row[0]
                data["success"] += row[1]
                data["status_code"].append(row[2])
                data["prompt_tokens"].append(row[3])
                data["completion_tokens"].append(row[4])
                data["e2e_time"].append(row[5])
                data["model"].append(row[6])
            return data
        finally:
            conn.close()

    def render_card_fresh(self, current_card, data):
        self._already_rendered = True
        current_card.clear()
        current_card.append(Markdown("## Metrics"))

        self._metrics_charts["request_success"] = BarPlot(
            title="Request success",
            category_name="category",
            value_name="amount",
            orientation="horizontal",
        )
        self._metrics_charts["latency_distribution"] = ViolinPlot(
            title="Latency distribution (s)",
            category_col_name="model",
            value_col_name="e2e_time",
        )

        current_card.append(
            Table(
                data=[
                    [
                        self._metrics_charts["request_success"],
                    ],
                    [self._metrics_charts["latency_distribution"]],
                ]
            )
        )
        current_card.refresh()

    def on_startup(self, current_card):
        current_card.append(Markdown("# Task-level NIM API metrics"))
        current_card.append(
            Markdown(
                "_waiting for data to appear_",
            )
        )
        current_card.refresh()

    def on_error(self, current_card, error_message):

        if isinstance(error_message, FileNotFoundError):
            return

        if not self._already_rendered:
            current_card.clear()
            current_card.append(
                Markdown(
                    f"## Error: {str(error_message)}",
                )
            )
            current_card.refresh()

    def update_only_components(self, current_card, data_object):

        # update request success data
        self._metrics_charts["request_success"].spec["data"][0]["values"] = [
            {
                "category": "Successful requests",
                "amount": data_object["metrics"]["success"],
            },
            {"category": "Errors", "amount": data_object["metrics"]["error"]},
        ]

        latency_data = []
        times = []
        for m, e in zip(
            data_object["metrics"]["model"], data_object["metrics"]["e2e_time"]
        ):
            latency_data.append({"model": m, "e2e_time": e})
            times.append(e)

        # update latency data
        self._metrics_charts["latency_distribution"].spec["data"][0][
            "values"
        ] = latency_data

        # update domain for latency plot
        min_time = min(times)
        max_time = max(times)
        for scale in self._metrics_charts["latency_distribution"].spec["scales"]:
            if scale["name"] == "xscale":
                scale["domain"] = [min_time - max_time * 0.1, max_time + max_time * 0.1]

        current_card.refresh()

    def on_update(self, current_card, data_object):
        data_object_keys = set(data_object.keys())
        if len(data_object_keys) == 0:
            return
        if len(self._metrics_charts) == 0:
            self.render_card_fresh(current_card, data_object)
            return
        elif len(data_object["metrics"]["status_code"]) == 0:
            return
        else:
            self.update_only_components(current_card, data_object)
            return



import os
import sys
import time
import subprocess
from io import BytesIO
from datetime import datetime, timezone

from metaflow.exception import MetaflowException


def kill_process_and_descendants(pid, termination_timeout=1, iterations=20, delay=0.5):
    for i in range(iterations):
        try:
            subprocess.check_call(["pkill", "-TERM", "-P", str(pid)])
            subprocess.check_call(["kill", "-TERM", str(pid)])

            time.sleep(termination_timeout)

            subprocess.check_call(["pkill", "-KILL", "-P", str(pid)])
            subprocess.check_call(["kill", "-KILL", str(pid)])
        except subprocess.CalledProcessError:
            pass

        # Don't delay after the last iteration
        if i < iterations - 1:
            time.sleep(delay)


class HeartbeatStore(object):
    def __init__(
        self,
        main_pid=None,
        storage_backend=None,
        emit_frequency=30,
        missed_heartbeat_timeout=60,
        monitor_frequency=15,
        max_missed_heartbeats=3,
    ) -> None:
        self.main_pid = main_pid
        self.storage_backend = storage_backend
        self.emit_frequency = emit_frequency
        self.monitor_frequency = monitor_frequency
        self.missed_heartbeat_timeout = missed_heartbeat_timeout
        self.max_missed_heartbeats = max_missed_heartbeats
        self.missed_heartbeats = 0

    def emit_heartbeat(self, heartbeat_prefix: str, folder_name=None):
        heartbeat_key = f"{heartbeat_prefix}/heartbeat"
        if folder_name:
            heartbeat_key = f"{folder_name}/{heartbeat_key}"

        while True:
            try:
                epoch_string = str(datetime.now(timezone.utc).timestamp()).encode(
                    "utf-8"
                )
                self.storage_backend.save_bytes(
                    [(heartbeat_key, BytesIO(bytes(epoch_string)))], overwrite=True
                )
            except Exception as e:
                print(f"Error writing heartbeat: {e}")
                sys.exit(1)

            time.sleep(self.emit_frequency)

    def emit_tombstone(self, tombstone_prefix: str, folder_name=None):
        tombstone_key = f"{tombstone_prefix}/tombstone"
        if folder_name:
            tombstone_key = f"{folder_name}/{tombstone_key}"

        tombstone_string = "tombstone".encode("utf-8")
        try:
            self.storage_backend.save_bytes(
                [(tombstone_key, BytesIO(bytes(tombstone_string)))], overwrite=True
            )
        except Exception as e:
            print(f"Error writing tombstone: {e}")
            sys.exit(1)

    def __handle_tombstone(self, path):
        if path is not None:
            with open(path) as f:
                contents = f.read()
                if "tombstone" in contents:
                    print("[Outerbounds] Tombstone detected. Terminating the task..")
                    kill_process_and_descendants(self.main_pid)
                    sys.exit(1)

    def __handle_heartbeat(self, path):
        if path is not None:
            with open(path) as f:
                contents = f.read()
                current_timestamp = datetime.now(timezone.utc).timestamp()
                if current_timestamp - float(contents) > self.missed_heartbeat_timeout:
                    self.missed_heartbeats += 1
                else:
                    self.missed_heartbeats = 0
        else:
            self.missed_heartbeats += 1

        if self.missed_heartbeats > self.max_missed_heartbeats:
            print(
                f"[Outerbounds] Missed {self.max_missed_heartbeats} consecutive heartbeats. Terminating the task.."
            )
            kill_process_and_descendants(self.main_pid)
            sys.exit(1)

    def is_main_process_running(self):
        try:
            # Check if the process is running
            os.kill(self.main_pid, 0)
        except ProcessLookupError:
            return False
        return True

    def monitor(self, heartbeat_prefix: str, tombstone_prefix: str, folder_name=None):
        heartbeat_key = f"{heartbeat_prefix}/heartbeat"
        if folder_name:
            heartbeat_key = f"{folder_name}/{heartbeat_key}"

        tombstone_key = f"{tombstone_prefix}/tombstone"
        if folder_name:
            tombstone_key = f"{folder_name}/{tombstone_key}"

        while self.is_main_process_running():
            with self.storage_backend.load_bytes(
                [heartbeat_key, tombstone_key]
            ) as results:
                for key, path, _ in results:
                    if key == tombstone_key:
                        self.__handle_tombstone(path)
                    elif key == heartbeat_key:
                        self.__handle_heartbeat(path)

            time.sleep(self.monitor_frequency)


if __name__ == "__main__":
    from metaflow.plugins import DATASTORES
    from metaflow.metaflow_config import NVIDIA_HEARTBEAT_THRESHOLD

    if len(sys.argv) != 4:
        print("Usage: heartbeat_store.py <main_pid> <datastore_type> <folder_name>")
        sys.exit(1)
    _, main_pid, datastore_type, folder_name = sys.argv

    if datastore_type not in ("azure", "gs", "s3"):
        print(f"Datastore unsupported for type: {datastore_type}")
        sys.exit(1)

    datastores = [d for d in DATASTORES if d.TYPE == datastore_type]
    datastore_sysroot = datastores[0].get_datastore_root_from_config(
        lambda *args, **kwargs: None
    )
    if datastore_sysroot is None:
        raise MetaflowException(
            msg="METAFLOW_DATASTORE_SYSROOT_{datastore_type} must be set!".format(
                datastore_type=datastore_type.upper()
            )
        )

    storage = datastores[0](datastore_sysroot)

    heartbeat_prefix = f"{os.getenv('MF_PATHSPEC')}/{os.getenv('MF_ATTEMPT')}"
    flow_name, run_id, _, _ = os.getenv("MF_PATHSPEC").split("/")
    tombstone_prefix = f"{flow_name}/{run_id}"

    store = HeartbeatStore(
        main_pid=int(main_pid),
        storage_backend=storage,
        max_missed_heartbeats=int(NVIDIA_HEARTBEAT_THRESHOLD),
    )

    store.monitor(
        heartbeat_prefix=heartbeat_prefix,
        tombstone_prefix=tombstone_prefix,
        folder_name=folder_name,
    )



# These two fields will show up within `metaflow_version` task metadata.
# Setting to major version of ob-metaflow-extensions only, so we don't keep trying
# (and failing) to keep this in sync with setup.py
# E.g. "2.7.22.1+ob(v1)"
__version__ = "v1"
__mf_extensions__ = "ob"


# Must match the signature of metaflow.plugins.aws.aws_client.get_aws_client
# This function is called by the "userland" code inside tasks. Metaflow internals
# will call the function in metaflow.plugins.aws.aws_client.get_aws_client directly.
#
# Unlike the original function, this wrapper will use the CSPR role if both of the following
# conditions are met:
#
#  1. CSPR is set
#  2. user didn't provide a role to assume explicitly.
#
def get_aws_client(
    module, with_error=False, role_arn=None, session_vars=None, client_params=None
):
    import metaflow.plugins.aws.aws_client

    from metaflow_extensions.outerbounds.plugins import USE_CSPR_ROLE_ARN_IF_SET

    return metaflow.plugins.aws.aws_client.get_aws_client(
        module,
        with_error=with_error,
        role_arn=role_arn or USE_CSPR_ROLE_ARN_IF_SET,
        session_vars=session_vars,
        client_params=client_params,
    )


# This should match the signature of metaflow.plugins.datatools.s3.S3.
#
# This assumes that "userland" code inside tasks will call this, while Metaflow
# internals will call metaflow.plugins.datatools.s3.S3 directly.
#
# This wrapper will make S3() use the CSPR role if its set, and user didn't provide
# a role to assume explicitly.
def S3(*args, **kwargs):
    import sys
    import metaflow.plugins.datatools.s3
    from metaflow_extensions.outerbounds.plugins import USE_CSPR_ROLE_ARN_IF_SET

    if "role" not in kwargs or kwargs["role"] is None:
        kwargs["role"] = USE_CSPR_ROLE_ARN_IF_SET

    return metaflow.plugins.datatools.s3.S3(*args, **kwargs)


from .. import profilers
from ..plugins.snowflake import Snowflake



import os
from metaflow.cards import (
    Markdown,
    Table,
    VegaChart,
    ProgressBar,
    MetaflowCardComponent,
    Artifact,
)
import math
from metaflow.plugins.cards.card_modules.components import (
    with_default_component_id,
    TaskToDict,
    ArtifactsComponent,
    render_safely,
)
import datetime
from metaflow.metaflow_current import current
import json
from functools import wraps
from collections import defaultdict
from threading import Thread, Event
import time

DEFAULT_WIDTH = 500
DEFAULT_HEIGHT = 200
DEFAULT_PADDING = 10
BG_COLOR = "#f2eeea"  # sand-100
VIEW_FILL = "#faf7f4"  # sand-200
GREYS = [
    "#ebe8e5",
    "#b2afac",
    "#6a6867",
]
BLACK = "#31302f"
GREENS = ["#dae8e2", "#3e8265", "#4c9878", "#428a6b", "#37795d"]
YELLOWS = ["#faf1db", "#f7e2b1", "#fbd784", "#e4b957", "#d7a530"]
PURPLES = ["#f5eff9", "#e7d4f3", "#976bac", "#8e53a9", "#77458f"]
REDS = ["#fce5e2", "#f3b6af", "#e6786c", "#e35f50", "#ce493a"]
BLUES = ["#dfe9f4", "#bdd8f2", "#88b7e3", "#6799c8", "#4e7ca7"]
ALL_COLORS = [
    # GREENS[0], PURPLES[0], REDS[0], BLUES[0], YELLOWS[0],
    GREENS[1],
    PURPLES[1],
    REDS[1],
    BLUES[1],
    YELLOWS[1],
    GREENS[2],
    PURPLES[2],
    REDS[2],
    BLUES[2],
    YELLOWS[2],
    GREENS[3],
    PURPLES[3],
    REDS[3],
    BLUES[3],
    YELLOWS[3],
    GREENS[4],
    PURPLES[4],
    REDS[4],
    BLUES[4],
    YELLOWS[4],
]


def update_spec_data(spec, data):
    spec["data"]["values"].append(data)
    return spec


def update_data_object(data_object, data):
    data_object["values"].append(data)
    return data_object


def line_chart_spec(
    title=None,
    category_name="u",
    y_name="v",
    xtitle=None,
    ytitle=None,
    width=DEFAULT_WIDTH,
    height=DEFAULT_HEIGHT,
    with_params=True,
    x_axis_temporal=False,
):
    parameters = [
        {
            "name": "interpolate",
            "value": "linear",
            "bind": {
                "input": "select",
                "options": [
                    "basis",
                    "cardinal",
                    "catmull-rom",
                    "linear",
                    "monotone",
                    "natural",
                    "step",
                    "step-after",
                    "step-before",
                ],
            },
        },
        {
            "name": "tension",
            "value": 0,
            "bind": {"input": "range", "min": 0, "max": 1, "step": 0.05},
        },
        {
            "name": "strokeWidth",
            "value": 2,
            "bind": {"input": "range", "min": 0, "max": 10, "step": 0.5},
        },
        {
            "name": "strokeCap",
            "value": "butt",
            "bind": {"input": "select", "options": ["butt", "round", "square"]},
        },
        {
            "name": "strokeDash",
            "value": [1, 0],
            "bind": {
                "input": "select",
                "options": [[1, 0], [8, 8], [8, 4], [4, 4], [4, 2], [2, 1], [1, 1]],
            },
        },
    ]
    parameter_marks = {
        "interpolate": {"expr": "interpolate"},
        "tension": {"expr": "tension"},
        "strokeWidth": {"expr": "strokeWidth"},
        "strokeDash": {"expr": "strokeDash"},
        "strokeCap": {"expr": "strokeCap"},
    }
    spec = {
        "title": title if title else "Line Chart",
        "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
        "width": DEFAULT_WIDTH,
        "height": DEFAULT_HEIGHT,
        "background": BG_COLOR,
        "padding": DEFAULT_PADDING,
        "view": {"fill": VIEW_FILL},
        "params": parameters if with_params else [],
        "data": {"name": "values", "values": []},
        "mark": {
            "type": "line",
            "tooltip": True,
            **(parameter_marks if with_params else {}),
        },
        "selection": {"grid": {"type": "interval", "bind": "scales"}},
        "encoding": {
            "x": {
                "field": category_name,
                "title": xtitle if xtitle else category_name,
                **({"timeUnit": "seconds"} if x_axis_temporal else {}),
                **({"type": "quantitative"} if not x_axis_temporal else {}),
            },
            "y": {
                "field": y_name,
                "type": "quantitative",
                "title": ytitle if ytitle else y_name,
            },
        },
    }
    data = {"values": []}
    return spec, data


class LineChart(MetaflowCardComponent):
    REALTIME_UPDATABLE = True

    def __init__(
        self,
        title,
        xtitle,
        ytitle,
        category_name,
        y_name,
        with_params=False,
        x_axis_temporal=False,
    ):
        super().__init__()

        self.spec, _ = line_chart_spec(
            title=title,
            xtitle=xtitle,
            ytitle=ytitle,
            category_name=category_name,
            y_name=y_name,
            with_params=with_params,
            x_axis_temporal=x_axis_temporal,
        )

    def update(self, data):  # Can take a diff
        self.spec = update_spec_data(self.spec, data)

    @with_default_component_id
    def render(self):
        vega_chart = VegaChart(self.spec, show_controls=True)
        vega_chart.component_id = self.component_id
        return vega_chart.render()


class ArtifactTable(Artifact):
    def __init__(self, data_dict):
        self._data = data_dict
        self._task_to_dict = TaskToDict(only_repr=True)

    @with_default_component_id
    @render_safely
    def render(self):
        _art_list = []
        for k, v in self._data.items():
            _art = self._task_to_dict.infer_object(v)
            _art["name"] = k
            _art_list.append(_art)

        af_component = ArtifactsComponent(data=_art_list)
        af_component.component_id = self.component_id
        return af_component.render()


# fmt: off
class BarPlot(MetaflowCardComponent):
    REALTIME_UPDATABLE = True

    def __init__(self, title, category_name, value_name, orientation="vertical"):

        if orientation not in ["vertical", "horizontal"]:
            raise ValueError("orientation must be either 'vertical' or 'horizontal'")

        super().__init__()
        self.spec = {
            "title": title,
            "$schema": "https://vega.github.io/schema/vega/v5.json",
            "description": "A basic bar chart example to show a count of values grouped by a category.",
            "background": BG_COLOR,
            "view": {"fill": VIEW_FILL},
            "width": DEFAULT_WIDTH,
            "height": DEFAULT_HEIGHT,
            "padding": DEFAULT_PADDING,
            "data": [{"name": "table", "values": []}],
            "signals": [
                {
                    "name": "tooltip",
                    "value": {},
                    "on": [
                        {"events": "rect:pointerover", "update": "datum"},
                        {"events": "rect:pointerout", "update": "{}"},
                    ],
                }
            ],
            "scales": [
                {
                    "name": "xscale" if orientation == "vertical" else "yscale",
                    "type": "band",
                    "domain": {"data": "table", "field": category_name},
                    "range": "width" if orientation == "vertical" else "height",
                    "padding": 0.25,
                    "round": True,
                },
                {
                    "name": "yscale" if orientation == "vertical" else "xscale",
                    "domain": {"data": "table", "field": value_name},
                    "nice": True,
                    "range": "height" if orientation == "vertical" else "width",
                },
                {
                    "name": "color",
                    "type": "ordinal",
                    "domain": {"data": "table", "field": category_name},
                    "range": ALL_COLORS,
                },
            ],
            "axes": [
                {"orient": "bottom", "scale": "xscale", "zindex": 1},
                {"orient": "left", "scale": "yscale", "zindex": 1},
            ],
            "marks": [
                {
                    "type": "rect",
                    "from": {"data": "table"},
                    "encode": {
                        "enter": {
                            "x": {
                                "scale": "xscale",
                                "field": (
                                    category_name
                                    if orientation == "vertical"
                                    else value_name
                                ),
                            },
                            "y": {
                                "scale": "yscale",
                                "field": (
                                    value_name
                                    if orientation == "vertical"
                                    else category_name
                                ),
                            },
                            f"{'y2' if orientation == 'vertical' else 'x2'}": {
                                "scale": (
                                    "yscale" if orientation == "vertical" else "xscale"
                                ),
                                "value": 0,
                            },
                            "width": {"scale": "xscale", "band": 1},
                            "height": {"scale": "yscale", "band": 1},
                        },
                        "update": {
                            "fill": {"value": GREENS[0]},
                        },
                        "hover": {"fill": {"value": GREENS[2]}},
                    },
                },
                {
                    "type": "text",
                    "encode": {
                        "enter": {
                            "align": {"value": "center"},
                            "baseline": {"value": "bottom"},
                            "fill": {"value": BG_COLOR},
                        },
                        "update": {
                            "x": {
                                "scale": "xscale",
                                "signal": f"tooltip.{category_name if orientation == 'vertical' else value_name}",
                                f"{'band' if orientation == 'vertical' else 'offset'}": (
                                    0.5 if orientation == "vertical" else -10
                                ),
                            },
                            "y": {
                                "scale": "yscale",
                                "signal": f"tooltip.{value_name if orientation == 'vertical' else category_name}",
                                f"{'band' if orientation == 'horizontal' else 'offset'}": (
                                    0.5 if orientation == "horizontal" else 20
                                ),
                            },
                            "text": {"signal": f"tooltip.{value_name}"},
                            "fillOpacity": [
                                {"test": "datum === tooltip", "value": 0},
                                {"value": 1},
                            ],
                        },
                    },
                },
            ],
        }

    def update(self, data):  # Can take a diff
        self.spec = update_spec_data(self.spec, data)

    @with_default_component_id
    def render(self):
        vega_chart = VegaChart(self.spec, show_controls=True)
        vega_chart.component_id = self.component_id
        return vega_chart.render()


class ViolinPlot(MetaflowCardComponent):
    REALTIME_UPDATABLE = True

    def __init__(self, title, category_col_name, value_col_name):
        super().__init__()

        self.spec = {
            "title": title,
            "$schema": "https://vega.github.io/schema/vega/v5.json",
            "description": "A violin chart to show a distributional properties of each category.",
            "background": BG_COLOR,
            "view": {"fill": VIEW_FILL},
            "width": DEFAULT_WIDTH,
            "height": DEFAULT_HEIGHT,
            "padding": DEFAULT_PADDING,
            "config": {
                "axisBand": {"bandPosition": 1, "tickExtra": True, "tickOffset": 0}
            },
            "signals": [
                {"name": "plotWidth", "value": 75},
                {"name": "height", "update": "(plotWidth + 10) * 3"},
                {
                    "name": "bandwidth",
                    "value": 0.1,
                    "bind": {"input": "range", "min": 0, "max": 0.2, "step": 0.01},
                },
            ],
            "data": [
                {"name": "src", "values": []},
                {
                    "name": "density",
                    "source": "src",
                    "transform": [
                        {
                            "type": "kde",
                            "groupby": [category_col_name],
                            "field": value_col_name,
                            "bandwidth": {"signal": "bandwidth"},
                            "extent": {"signal": "domain('xscale')"},
                        }
                    ],
                },
                {
                    "name": "stats",
                    "source": "src",
                    "transform": [
                        {
                            "type": "aggregate",
                            "groupby": [category_col_name],
                            "fields": [value_col_name, value_col_name, value_col_name],
                            "ops": ["q1", "q3", "median"],
                            "as": ["q1", "q3", "median"],
                        }
                    ],
                },
            ],
            "scales": [
                {
                    "name": "layout",
                    "type": "band",
                    "range": "height",
                    "domain": {"data": "src", "field": category_col_name},
                },
                {
                    "name": "xscale",
                    "type": "linear",
                    "range": "width",
                    "round": True,
                    "domain": {"data": "src", "field": value_col_name},
                    "zero": False,
                    "nice": True,
                },
                {
                    "name": "hscale",
                    "type": "linear",
                    "range": [0, {"signal": "plotWidth"}],
                    "domain": {"data": "density", "field": "density"},
                },
                {
                    "name": "color",
                    "type": "ordinal",
                    "domain": {"data": "src", "field": category_col_name},
                    "range": ALL_COLORS,
                },
            ],
            "axes": [
                {"orient": "bottom", "scale": "xscale", "zindex": 1},
                {"orient": "left", "scale": "layout", "zindex": 1},
            ],
            "marks": [
                {
                    "type": "group",
                    "from": {
                        "facet": {
                            "data": "density",
                            "name": "violin",
                            "groupby": category_col_name,
                        }
                    },
                    "encode": {
                        "enter": {
                            "yc": {
                                "scale": "layout",
                                "field": category_col_name,
                                "band": 0.5,
                            },
                            "height": {"signal": "plotWidth"},
                            "width": {"signal": "width"},
                        }
                    },
                    "data": [
                        {
                            "name": "summary",
                            "source": "stats",
                            "transform": [
                                {
                                    "type": "filter",
                                    "expr": f"datum.{category_col_name} === parent.{category_col_name}",
                                }
                            ],
                        }
                    ],
                    "marks": [
                        {
                            "type": "area",
                            "from": {"data": "violin"},
                            "encode": {
                                "enter": {
                                    "fill": {
                                        "scale": "color",
                                        "field": {"parent": category_col_name},
                                    }
                                },
                                "update": {
                                    "x": {"scale": "xscale", "field": "value"},
                                    "yc": {"signal": "plotWidth / 2"},
                                    "height": {"scale": "hscale", "field": "density"},
                                },
                            },
                        },
                        {
                            "type": "rect",
                            "from": {"data": "summary"},
                            "encode": {
                                "enter": {
                                    "fill": {"value": BLACK},
                                    "height": {"value": 2},
                                },
                                "update": {
                                    "x": {"scale": "xscale", "field": "q1"},
                                    "x2": {"scale": "xscale", "field": "q3"},
                                    "yc": {"signal": "plotWidth / 2"},
                                },
                            },
                        },
                        {
                            "type": "rect",
                            "from": {"data": "summary"},
                            "encode": {
                                "enter": {
                                    "fill": {"value": BLACK},
                                    "width": {"value": 2},
                                    "height": {"value": 8},
                                },
                                "update": {
                                    "x": {"scale": "xscale", "field": "median"},
                                    "yc": {"signal": "plotWidth / 2"},
                                },
                            },
                        },
                    ],
                }
            ],
        }

    def update(self, data):  # Can take a diff
        self.spec = update_spec_data(self.spec, data)

    @with_default_component_id
    def render(self):
        vega_chart = VegaChart(self.spec, show_controls=True)
        vega_chart.component_id = self.component_id
        return vega_chart.render()
# fmt: on



import json
import os
import sys
from pathlib import Path
from typing import Optional, Dict

import requests
from metaflow.exception import MetaflowException
from requests.models import HTTPError
from metaflow_extensions.outerbounds.plugins.perimeters import (
    get_perimeter_config_url_if_set_in_ob_config,
)

OBP_REMOTE_CONFIG_KEY = "OBP_METAFLOW_CONFIG_URL"
HOSTNAME_KEY = "OBP_API_SERVER"
AUTH_KEY = "METAFLOW_SERVICE_AUTH_KEY"
PERIMETER_KEY = "OBP_PERIMETER"
CONFIG_READ_ONCE_KEY = "__REMOTE_CONFIG_HAS_BEEN_RESOLVED__"


def read_config_from_local() -> Optional[Path]:
    default_path = Path.home() / ".metaflowconfig"
    home = Path(os.environ.get("METAFLOW_HOME", default_path))

    profile = os.environ.get("METAFLOW_PROFILE")
    config_path = home / f"config_{profile}.json" if profile else home / "config.json"

    if config_path.exists() and config_path.is_file():
        _init_debug(f"using config from {config_path}")
        return config_path

    # we should error because the user wants a specific config
    if profile:
        raise MetaflowException(
            f"Unable to locate METAFLOW_PROFILE {profile} in {config_path}"
        )

    # there's no config and that's ok. Metaflow uses environment variables as its primary way to set values
    # and will fallback to local settings if no config is present
    _init_debug(f"no config present at path {config_path}")
    return None


def resolve_config_from_remote(remote_url: str, auth_token: str) -> Dict[str, str]:
    _init_debug(f"retrieving config from {remote_url}")

    headers = {"x-api-key": auth_token}
    try:
        response = requests.get(remote_url, headers=headers)
        _init_debug(
            f"response\nstatus code: {response.status_code}\nbody: {response.text}"
        )

        response.raise_for_status()
        data = response.json()
        return data["config"]
    except HTTPError:
        raise MetaflowException(
            "Error fetching resolving configuration. Make sure you have run \
                                `outerbounds configure` with the correct value"
        )


def init_config() -> Dict[str, str]:
    """
    OSS Metaflow reads the config file on every step initialization. This is because OSS assumes config files change
    relatively infrequently. We want to avoid config values changing between flow steps. Our solution to prevent this
    is to read a config once and cache it on an environment variable. Environment variables carry over between steps
    because steps are executed in subprocesses (local) or environments which expect environment variables to be set.
    """
    _init_debug("starting initialization")
    config_json = os.environ.get(CONFIG_READ_ONCE_KEY)
    if config_json:
        _init_debug("reading config from environment")
        return json.loads(config_json)

    config_path = read_config_from_local()
    if not config_path:
        return {}

    try:
        remote_config = json.loads(config_path.read_text())
    except ValueError:
        raise MetaflowException(
            "Error decoding your metaflow config. Please run the `outerbounds configure` \
                                 command with the string provided in the Outerbounds dashboard"
        )

    perimeter_config_url = get_perimeter_config_url_if_set_in_ob_config()
    if perimeter_config_url:
        remote_config[OBP_REMOTE_CONFIG_KEY] = perimeter_config_url

    # users still have a legacy format and that's ok.
    if OBP_REMOTE_CONFIG_KEY not in remote_config:
        return remote_config

    metaflow_config = resolve_config_from_remote(
        remote_url=remote_config[OBP_REMOTE_CONFIG_KEY],
        auth_token=remote_config[AUTH_KEY],
    )

    # set cache
    os.environ[CONFIG_READ_ONCE_KEY] = json.dumps(metaflow_config)
    return metaflow_config


DEBUG_CONFIG = os.environ.get("METAFLOW_DEBUG_CONFIG")


def _init_debug(*args, **kwargs):
    if DEBUG_CONFIG:
        init_str = "ob_extension_init:"
        kwargs["file"] = sys.stderr
        print(init_str, *args, **kwargs)



import os
import sys
import time
import traceback

from metaflow._vendor import click
from metaflow import util
from metaflow import R
from metaflow.exception import METAFLOW_EXIT_DISALLOW_RETRY
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.metadata_provider.util import sync_local_metadata_from_datastore
from metaflow.mflog import TASK_LOG_SOURCE

from .snowpark_exceptions import SnowparkKilledException
from .snowpark import Snowpark


@click.group()
def cli():
    pass


@cli.group(help="Commands related to Snowpark.")
def snowpark():
    pass


@snowpark.command(
    help="Execute a single task using Snowpark. This command calls the "
    "top-level step command inside a Snowpark job with the given options. "
    "Typically you do not call this command directly; it is used internally by "
    "Metaflow."
)
@click.argument("step-name")
@click.argument("code-package-sha")
@click.argument("code-package-url")
@click.option(
    "--executable", help="Executable requirement for Snowpark Container Services."
)
# below are snowpark specific options
@click.option(
    "--account",
    help="Account ID for Snowpark Container Services.",
)
@click.option(
    "--user",
    help="Username for Snowpark Container Services.",
)
@click.option(
    "--password",
    help="Password for Snowpark Container Services.",
)
@click.option(
    "--role",
    help="Role for Snowpark Container Services.",
)
@click.option(
    "--database",
    help="Database for Snowpark Container Services.",
)
@click.option(
    "--warehouse",
    help="Warehouse for Snowpark Container Services.",
)
@click.option(
    "--schema",
    help="Schema for Snowpark Container Services.",
)
@click.option(
    "--image",
    help="Docker image requirement for Snowpark Container Services. In name:version format.",
)
@click.option("--stage", help="Stage requirement for Snowpark Container Services.")
@click.option(
    "--compute-pool", help="Compute Pool requirement for Snowpark Container Services."
)
@click.option("--volume-mounts", multiple=True)
@click.option(
    "--external-integration",
    multiple=True,
    help="External Integration requirement for Snowpark Container Services.",
)
@click.option("--cpu", help="CPU requirement for Snowpark Container Services.")
@click.option("--gpu", help="GPU requirement for Snowpark Container Services.")
@click.option("--memory", help="Memory requirement for Snowpark Container Services.")
# TODO: secrets, volumes for snowpark
# TODO: others to consider: ubf-context, num-parallel
@click.option("--run-id", help="Passed to the top-level 'step'.")
@click.option("--task-id", help="Passed to the top-level 'step'.")
@click.option("--input-paths", help="Passed to the top-level 'step'.")
@click.option("--split-index", help="Passed to the top-level 'step'.")
@click.option("--clone-path", help="Passed to the top-level 'step'.")
@click.option("--clone-run-id", help="Passed to the top-level 'step'.")
@click.option(
    "--tag", multiple=True, default=None, help="Passed to the top-level 'step'."
)
@click.option("--namespace", default=None, help="Passed to the top-level 'step'.")
@click.option("--retry-count", default=0, help="Passed to the top-level 'step'.")
@click.option(
    "--max-user-code-retries", default=0, help="Passed to the top-level 'step'."
)
# TODO: this is not used anywhere as of now...
@click.option(
    "--run-time-limit",
    default=5 * 24 * 60 * 60,  # Default is set to 5 days
    help="Run time limit in seconds for Snowpark container.",
)
@click.pass_context
def step(
    ctx,
    step_name,
    code_package_sha,
    code_package_url,
    executable=None,
    account=None,
    user=None,
    password=None,
    role=None,
    database=None,
    warehouse=None,
    schema=None,
    image=None,
    stage=None,
    compute_pool=None,
    volume_mounts=None,
    external_integration=None,
    cpu=None,
    gpu=None,
    memory=None,
    run_time_limit=None,
    **kwargs
):
    def echo(msg, stream="stderr", job_id=None, **kwargs):
        msg = util.to_unicode(msg)
        if job_id:
            msg = "[%s] %s" % (job_id, msg)
        ctx.obj.echo_always(msg, err=(stream == sys.stderr), **kwargs)

    if R.use_r():
        entrypoint = R.entrypoint()
    else:
        executable = ctx.obj.environment.executable(step_name, executable)
        entrypoint = "%s -u %s" % (executable, os.path.basename(sys.argv[0]))

    top_args = " ".join(util.dict_to_cli_options(ctx.parent.parent.params))

    input_paths = kwargs.get("input_paths")
    split_vars = None
    if input_paths:
        max_size = 30 * 1024
        split_vars = {
            "METAFLOW_INPUT_PATHS_%d" % (i // max_size): input_paths[i : i + max_size]
            for i in range(0, len(input_paths), max_size)
        }
        kwargs["input_paths"] = "".join("${%s}" % s for s in split_vars.keys())

    step_args = " ".join(util.dict_to_cli_options(kwargs))

    step_cli = "{entrypoint} {top_args} step {step} {step_args}".format(
        entrypoint=entrypoint,
        top_args=top_args,
        step=step_name,
        step_args=step_args,
    )
    node = ctx.obj.graph[step_name]

    # Get retry information
    retry_count = kwargs.get("retry_count", 0)
    retry_deco = [deco for deco in node.decorators if deco.name == "retry"]
    minutes_between_retries = None
    if retry_deco:
        minutes_between_retries = int(
            retry_deco[0].attributes.get("minutes_between_retries", 1)
        )
    if retry_count:
        ctx.obj.echo_always(
            "Sleeping %d minutes before the next retry" % minutes_between_retries
        )
        time.sleep(minutes_between_retries * 60)

    # Set batch attributes
    task_spec = {
        "flow_name": ctx.obj.flow.name,
        "step_name": step_name,
        "run_id": kwargs["run_id"],
        "task_id": kwargs["task_id"],
        "retry_count": str(retry_count),
    }
    attrs = {"metaflow.%s" % k: v for k, v in task_spec.items()}
    attrs["metaflow.user"] = util.get_username()
    attrs["metaflow.version"] = ctx.obj.environment.get_environment_info()[
        "metaflow_version"
    ]

    # Set environment
    env = {}
    env_deco = [deco for deco in node.decorators if deco.name == "environment"]
    if env_deco:
        env = env_deco[0].attributes["vars"]

    # Add the environment variables related to the input-paths argument
    if split_vars:
        env.update(split_vars)

    # Set log tailing.
    ds = ctx.obj.flow_datastore.get_task_datastore(
        mode="w",
        run_id=kwargs["run_id"],
        step_name=step_name,
        task_id=kwargs["task_id"],
        attempt=int(retry_count),
    )
    stdout_location = ds.get_log_location(TASK_LOG_SOURCE, "stdout")
    stderr_location = ds.get_log_location(TASK_LOG_SOURCE, "stderr")

    def _sync_metadata():
        if ctx.obj.metadata.TYPE == "local":
            sync_local_metadata_from_datastore(
                DATASTORE_LOCAL_DIR,
                ctx.obj.flow_datastore.get_task_datastore(
                    kwargs["run_id"], step_name, kwargs["task_id"]
                ),
            )

    try:
        snowpark = Snowpark(
            datastore=ctx.obj.flow_datastore,
            metadata=ctx.obj.metadata,
            environment=ctx.obj.environment,
            client_credentials={
                "account": account,
                "user": user,
                "password": password,
                "role": role,
                "database": database,
                "warehouse": warehouse,
                "schema": schema,
            },
        )
        with ctx.obj.monitor.measure("metaflow.snowpark.launch_job"):
            snowpark.launch_job(
                step_name=step_name,
                step_cli=step_cli,
                task_spec=task_spec,
                code_package_sha=code_package_sha,
                code_package_url=code_package_url,
                code_package_ds=ctx.obj.flow_datastore.TYPE,
                image=image,
                stage=stage,
                compute_pool=compute_pool,
                volume_mounts=volume_mounts,
                external_integration=external_integration,
                cpu=cpu,
                gpu=gpu,
                memory=memory,
                run_time_limit=run_time_limit,
                env=env,
                attrs=attrs,
            )
    except Exception:
        traceback.print_exc(chain=False)
        _sync_metadata()
        sys.exit(METAFLOW_EXIT_DISALLOW_RETRY)
    try:
        snowpark.wait(stdout_location, stderr_location, echo=echo)
    except SnowparkKilledException:
        # don't retry killed tasks
        traceback.print_exc()
        sys.exit(METAFLOW_EXIT_DISALLOW_RETRY)
    finally:
        _sync_metadata()
